/*
This is a Optical-Character-Recognition program
Copyright (C) 2000  Joerg Schulenburg

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 Joerg.Schulenburg@physik.uni-magdeburg.de
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "pgm2asc.h"
#include "gocr.h"
#ifdef HAVE_WCHAR_H
#include <wchar.h>
#endif

const wchar_t *getTextLine (int line) {
  int i;
  Element *elem;
  
  if (line < 0 || line > list_total(&(JOB->res.linelist)))
    return NULL;

  for ( i = 0, elem = JOB->res.linelist.header; i < line && elem != NULL; i++ )
    elem = elem->next;

  if ( elem != NULL )
    return (const wchar_t *)elem->data;

  return NULL;
}

void free_textlines(void) {
  for_each_data(&(JOB->res.linelist)) {
    if (list_get_current(&(JOB->res.linelist)))
      free(list_get_current(&(JOB->res.linelist)));
  } end_for_each(&(JOB->res.linelist));
  list_free(&(JOB->res.linelist));
}

/* append a char (c) to the string buffer (buffer) of length (len)
 * if buffer is to small or len==0 realloc buffer, len+=512
 */
wchar_t *append_to_line(wchar_t *buffer, wchar_t c, int *len) {
  wchar_t *t;
  int slen=0;
  if( (int)c == 0 ){
    fprintf(stderr,"\n#BUG: appending 0 to a line makes no sence!");
    return buffer;
  }
  if ( *len>0 ) slen= wcslen(buffer);
  if ( slen >= *len-1 ) {
    *len+=512;
    t = (wchar_t *)realloc(buffer, *len*sizeof(wchar_t));
    if( !t ) { fprintf(stderr,"realloc failed!\n"); *len-=512; return buffer; }
    else buffer = t;
  }
  t = buffer + slen;
#ifndef USE_UNICODE
  if( (char)c == 0 ) c='_';
#endif
  *t++ = c;
  *t = 0;
  return buffer;
}

/* collect all the chars from the box tree and write them to a string buffer
   mo is the mode: mode&8 means, use chars even if unsure recognized
 */
void store_boxtree_lines(int mo) {
  wchar_t *buffer;	/* temp buffer for text */
  int i = 0, j = 0, i1, dy=JOB->res.avY;
  int len = 1024;
  struct box *box2;

  buffer = (wchar_t *)malloc(len*sizeof(wchar_t));
  if ( !buffer ) { 
    fprintf(stderr,"malloc failed!\n"); 
    return;
  }
  *buffer = 0;

  if ( JOB->cfg.verbose&1 ) 
    fprintf(stderr,"# store boxtree to lines ...");

  for_each_data(&(JOB->res.boxlist)) {
    box2 = (struct box *)list_get_current(&(JOB->res.boxlist));
    if (box2->c == '\n') {
      dy=box2->m4-box2->m1; if (dy<7) dy=14;
      for (i = (box2->y1 - box2->y0) / (2 * dy) ; i > 0; i--) {
          buffer=append_to_line(buffer,'\n',&len);
          j++; // count chars in line
      }
      list_app( &(JOB->res.linelist), (void *)wcsdup(buffer) );
      memset(buffer, 0, len);
      j=0;  // reset counter for new line
    }
    if (box2->c == ' ')	// fill large gaps with spaces
    {
      if (JOB->res.avX) /* avoid SIGFPE */
      for (i = (box2->x1 - box2->x0) / (2 * JOB->res.avX) + 1; i > 0; i--) {
        buffer=append_to_line(buffer,' ',&len);
        j++;
      }
    }
    else if (box2->c != '\n') {
      if (j==0 && JOB->res.avX) /* first char in new line? */
      for (i = (box2->x0-JOB->res.lines.x0[box2->line]) / (2 * JOB->res.avX) + 1; i > 0; i--) {
        buffer=append_to_line(buffer,' ',&len); j++;
      }
      if (box2->obj) {
        for (i1=0;box2->obj[i1] && i1<1024;i1++) {
          buffer=append_to_line(buffer,(wchar_t)box2->obj[i1],&len);
          j++;
        }
        /* should we free box2->obj here and reset? not very clean */
        free(box2->obj); box2->obj=NULL;
      } else if (box2->c != UNKNOWN && box2->c!=0)
        buffer=append_to_line(buffer,box2->c,&len);
      else if (!(mo & 8))
        buffer=append_to_line(buffer,box2->ac,&len); // alternate char
      else
        buffer=append_to_line(buffer,UNKNOWN,&len);
      j++;
    }
    i++;
  } end_for_each(&(JOB->res.boxlist));

  /* do not forget last line */
  // is there no \n in the last line? If there is, delete next line.
  list_app( &(JOB->res.linelist), (void *)wcsdup(buffer) );
  free(buffer);
  if( JOB->cfg.verbose&1 ) fprintf(stderr,"... %d lines\n",i);
}
